#include "localdba.h"
#include "version.h"
#include "hsp.h"
#include "subseqhash.h"

#include "localcishit.h"

#include "hitlist.h"

boolean process_HSP(HSPset * set,Sequence * query,int query_pos,Sequence * tseq,SeqLookupResultStruct * res_struct,CompMat * mat)
{
  int k;

  for(k=0;k<set->len;k++) {
    if( on_HSP(set->hsp[k],query_pos,res_struct->pos) == TRUE ) {
      return FALSE;
    }
  }

  /* new HSP - extend and add */

  add_HSPset(set,new_HSP(NULL,query,tseq,query_pos,res_struct->pos,mat,5*Probability2Score(0.8/0.25)));

  return TRUE;
}

struct two_hspset {
  HSPset * forward;
  HSPset * reverse;
};

struct two_hspset query_to_reverse_target(Sequence * query,Sequence * target,DnaMatrix * dm,int qstart,int qend,int tstart,int tend)
{
  struct two_hspset out;
  HSPset * forward;
  HSPset * reverse;
  CompMat * cm;
  SeqLookupInterface * sli;
  SeqLookupClientInterface * slci;
  SeqLookupResultInterface * res;
  SeqLookupResultStruct * res_struct = NULL;
  Sequence * revseq;
  int i;
  int no;

  GTree * forward_tree;
  GTree * reverse_tree;

  cm = new_CompMat_from_DnaMatrix_flat(dm);

  /* reverse sequence */

  revseq = reverse_complement_Sequence(target);

  forward_tree = g_tree_new(g_int_equal);
  reverse_tree = g_tree_new(g_int_equal);




  /* build a hash based lookup table */

  sli = new_ghash_SeqLookupInterface();



  assert(cm);
  assert(target);
  assert(revseq);
  assert(sli);


  for(i=tstart;i<tend-7;i++) {
    if( i%1000 == 0 ) {
      info("Loaded %d positions in %s",i,target->name);
    }
    
    no = seq_number_dna_7mer_noN(target->seq+i);
    if( no != -1 ) {
      (*sli->add_direct_number)(sli->data,no,target,i);
    }

  }

  for(i=target->len-tend;i<target->len-tstart-7;i++) {
    if( i%1000 == 0 ) {
      info("Loaded %d positions in %s",i,target->name);
    }
    
    no = seq_number_dna_7mer_noN(revseq->seq+i);
    if( no != -1 ) {
      (*sli->add_direct_number)(sli->data,no,revseq,i);
    }

  }


  forward = HSPset_alloc_std();
  reverse = HSPset_alloc_std();

  /* scan the table with the query, testing whether new hits fit into HSP */

  slci = (*sli->get_client)(sli->data);

  for(i=qstart;i<qend-7;i++) {
    if( i%1000 == 0 ) {
      info("Considered %d positions in %s [forward %d HSPs, backward %d HSPs]\n",i,query->name,forward->len,reverse->len);
	}
    no = seq_number_dna_7mer_noN(query->seq+i);
    if( no == -1 ) {
      continue;
    }

    if( (*slci->is_populated)(sli->data,no) == FALSE) {
      continue;
    }

    res = (*slci->lookup)(sli->data,no);
    while( (*res->is_more)(res->data) ) {
      res_struct = (*res->next)(res->data,res_struct);
      if( res_struct->seq == target ) {
	process_HSP(forward,query,i,target,res_struct,cm);
      } else {
	process_HSP(reverse,query,i,revseq,res_struct,cm);
      }
    }
    free_SeqLookupResultInterface(res);
  } 

  free_SeqLookupClientInterface(slci);

  out.forward = forward;
  out.reverse = reverse;

  return out;
}


char * program_name = "promoterwise";


void show_version(FILE * ofp)
{
  fprintf(ofp,"%s\nVersion: %s\nReleased: %s\nCompiled: %s\n",program_name,VERSION_NUMBER,RELEASE_DAY,COMPILE_DATE);
  fprintf(ofp,"\nThis program is freely distributed under a Gnu Public License\n");
  fprintf(ofp,"The source code is copyright (c) EMBL and others\n");
  fprintf(ofp,"There is no warranty, implied or otherwise on the performance of this program\n");
  fprintf(ofp,"For more information read the GNULICENSE file in the distribution\n\n");
  fprintf(ofp,"Credits: Ewan Birney <birney@ebi.ac.uk>\n");
  exit(63);   
}

void show_help(FILE * ofp)
{
  fprintf(ofp,"%s query_sequence target_sequence\n",program_name);

  fprintf(ofp,"Seed restriction\n");
  fprintf(ofp,"  -s    query start position restriction\n");
  fprintf(ofp,"  -t    query end position restriction\n");
  fprintf(ofp,"  -u    target start position restriction\n");
  fprintf(ofp,"  -v    target end position restriction\n");
  show_help_LocalCisHitSetPara(ofp);
  show_help_HitListOutputImpl(ofp);
  show_help_DPRunImpl(ofp);

  show_standard_options(ofp);

}


int main(int argc,char ** argv)
{
  DPRunImpl * dpri = NULL;
  HitList * hl;
  HitListOutputImpl * hloi;

  Sequence * query;
  Sequence * target;
  Sequence * target_rev;
  struct two_hspset two;

  LocalCisHitSet * set;
  LocalCisHitSet * greedy_set;

  LocalCisHitScore * lchs;
  LocalCisHitSetPara * setpara;

  int qstart = -1;
  int qend   = -1;
  
  int tstart = -1;
  int tend   = -1;
  

  DnaMatrix * dm;
  DnaProbMatrix * dmp;

  dmp = DnaProbMatrix_from_match(0.75,NMaskType_BANNED);  
  assert(dmp);
  flat_null_DnaProbMatrix(dmp);  

  dm = DnaMatrix_from_DnaProbMatrix(dmp);
  
  dpri      = new_DPRunImpl_from_argv(&argc,argv);
  hloi      = new_HitListOutputImpl_from_argv(&argc,argv);
  setpara   = new_LocalCisHitSetPara_from_argv(&argc,argv);


  strip_out_integer_argument(&argc,argv,"s",&qstart);
  strip_out_integer_argument(&argc,argv,"t",&qend);
  strip_out_integer_argument(&argc,argv,"u",&tstart);
  strip_out_integer_argument(&argc,argv,"v",&tend);


  strip_out_standard_options(&argc,argv,show_help,show_version);
  if( argc != 3 ) {
    show_help(stdout);
    exit(12);
  }


  lchs = standard_LocalCisHitScore(NMaskType_VARIABLE);

  query = read_fasta_file_Sequence(argv[1]);
  target = read_fasta_file_Sequence(argv[2]);


  target_rev = reverse_complement_Sequence(target);

  assert(query);
  assert(target);


  if( qstart == -1 ) {
    qstart = 0;
  }
  if( qend == -1 ) {
    qend = query->len;
  }
  if( tstart == -1 ) {
    tstart = 0;
  }
  if( tend == -1 ) {
    tend = target->len;
  }

  
  two = query_to_reverse_target(query,target,dm,qstart,qend,tstart,tend);

  set = make_LocalCisHitSet(query,target,target_rev,two.forward,two.reverse,setpara,lchs,dpri);

  greedy_set = greedy_weed_LocalCisHitSet(set,setpara);


  hl = HitList_from_LocalCisHitSet(greedy_set);

  show_HitList_HitListOutputImpl(hloi,hl,stdout);

  return 0;
}



